from typing import List, Optional
from warnings import warn

from beam.base import ResourceSet
from beam.serializer import VolumeConfiguration
from beam.types import MountType


class MountResourceSet(ResourceSet):
    def PersistentVolume(
        self, name: str, path: str = "", app_path: Optional[str] = None, **_
    ):
        if app_path is not None:
            warn(
                "'app_path' is deprecated and will be removed in the future. "
                "Please use 'path' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.resources.append(
            VolumeConfiguration(
                name=name,
                local_path=None,
                app_path=path or app_path,
                mount_type=MountType.Persistent,
            )
        )

    def SharedVolume(
        self, name: str, path: str = "", app_path: Optional[str] = None, **_
    ):
        if app_path is not None:
            warn(
                "'app_path' is deprecated and will be removed in the future. "
                "Please use 'path' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.resources.append(
            VolumeConfiguration(
                name=name,
                local_path=None,
                app_path=path or app_path,
                mount_type=MountType.Shared,
            )
        )

    def from_config(self, mounts: List[dict]):
        if mounts is None:
            return

        for m in mounts:
            if m.get("mount_type") == MountType.Persistent:
                self.PersistentVolume(**m)

            elif m.get("mount_type") == MountType.Shared:
                self.SharedVolume(**m)



import json

from typing import List, Union
from beam.base import AbstractDataLoader
from beam.configs.resourceset import (
    TriggerResourceSet,
    OutputResourceSet,
    MountResourceSet,
)
from beam.configs.resource import (
    AutoscalingResource,
)
from beam.serializer import AppSpecV1Configuration, AppSpecV2Configuration
from beam.types import PythonVersion, GpuType
from beam.utils.parse import compose_cpu, compose_memory, load_requirements_file
from beam.exceptions import BeamInvalidAppSpecVersion


class AppSpecVersion:
    V1 = "v1"
    V2 = "v2"


class App(AbstractDataLoader):
    @staticmethod
    def from_config(config: Union[dict, str]) -> "App":
        if isinstance(config, str):
            config = json.loads(config)

        app_spec_version = config.get("app_spec_version", AppSpecVersion.V1)
        if app_spec_version == "":
            app_spec_version = AppSpecVersion.V1

        if app_spec_version == AppSpecVersion.V1:
            app_config = config.get("app")
            triggers = config.get("triggers")
            outputs = config.get("outputs")
            mounts = config.get("mounts")
            autoscaling = config.get("autoscaling")

            app = AppV1(**app_config)
            app.Trigger.from_config(triggers)
            app.Output.from_config(outputs)
            app.Mount.from_config(mounts)
            app.Trigger.AutoScaling.from_config(autoscaling)

            return app

        elif app_spec_version == AppSpecVersion.V2:
            app_config = config.get("app")
            triggers = config.get("triggers")
            outputs = config.get("outputs")
            mounts = config.get("mounts")
            autoscaling = config.get("autoscaling")

            app = AppV2(**app_config)
            app.Trigger.from_config(triggers)
            app.Output.from_config(outputs)
            app.Mount.from_config(mounts)
            app.Trigger.AutoScaling.from_config(autoscaling)

            return app
        else:
            raise BeamInvalidAppSpecVersion(f"invalid version: {app_spec_version}")


class AppV1(AbstractDataLoader):
    def __init__(
        self,
        *,
        name: str,
        cpu: Union[str, int],
        memory: str,
        gpu: str = GpuType.NoGPU,
        python_version: PythonVersion = PythonVersion.Python38,
        python_packages: Union[str, List[str]] = [],
        workspace: str = "./",
    ) -> None:
        """
        Keyword Arguments:
            name: the unique identifier for your app
            cpu: total number of cpu cores available to your app
            memory: total amount of memory available to your app
                - in the format [Number][Mi|Gi], for example 12Gi or 250Mi
            (Optional) gpu: type of gpu device available to your app (e.g. any, T4, A10G). Defaults to "", which means no GPU will be available
            (Optional) python_version: version of python to run your app code
            (Optional) python_packages: list of python packages you want to install, or the path to a requirements.txt file
                - e.g. "torch" or "torch==1.12.0"
            (Optional) workspace: directory to continously sync to your remote container during development
        """

        if isinstance(python_packages, str):
            python_packages = load_requirements_file(python_packages)

        self.Spec = AppSpecV1Configuration(
            name=name,
            cpu=compose_cpu(cpu),
            gpu=gpu,
            memory=compose_memory(memory),
            python_version=python_version,
            python_packages=python_packages,
            workspace=workspace,
        )

        self.Trigger = TriggerResourceSet()
        self.Output = OutputResourceSet()
        self.Mount = MountResourceSet()

    def dumps(self):
        return json.dumps(
            {
                "app_spec_version": AppSpecVersion.V1,
                "app": self.Spec.validate_and_dump(),
                "triggers": self.Trigger.dumps(),
                "outputs": self.Output.dumps(),
                "mounts": self.Mount.dumps(),
                "autoscaling": self.Trigger.AutoScaling.dumps(),
            }
        )


class AppV2(AbstractDataLoader):
    def __init__(
        self,
        *,
        name: str,
        cpu: Union[str, int],
        memory: str,
        gpu: str = GpuType.NoGPU,
        python_version: PythonVersion = PythonVersion.Python38,
        python_packages: Union[str, List[str]] = [],
        commands: Union[str, List[str]] = [],
        workspace: str = "./",
    ) -> None:
        """
        Keyword Arguments:
            name: the unique identifier for your app
            cpu: total number of cpu cores available to your app
            memory: total amount of memory available to your app
                - in the format [Number][Mi|Gi], for example 12Gi or 250Mi
            (Optional) gpu: type of gpu device available to your app (e.g. any, T4, A10G). Defaults to "", which means no GPU will be available
            (Optional) python_version: version of python to run your app code
            (Optional) python_packages: list of python packages you want to install, or the path to a requirements.txt file
                - e.g. "torch" or "torch==1.12.0"
            (Optional) commands: list of shell commands you'd like to run as we build your runtime
                - e.g. "apt-get install -y ffmpeg"
            (Optional) workspace: directory to continously sync to your remote container during development
        """

        if isinstance(python_packages, str):
            python_packages = load_requirements_file(python_packages)

        self.Spec = AppSpecV2Configuration(
            name=name,
            cpu=compose_cpu(cpu),
            gpu=gpu,
            memory=compose_memory(memory),
            python_version=python_version,
            python_packages=python_packages,
            commands=commands,
            workspace=workspace,
        )

        self.Trigger = TriggerResourceSet()
        self.Mount = MountResourceSet()
        self.Output = OutputResourceSet()

    def dumps(self):
        dump_config = {
            "app_spec_version": AppSpecVersion.V2,
            "app": self.Spec.validate_and_dump(),
            "triggers": self.Trigger.dumps(),
            "mounts": self.Mount.dumps(),
            "outputs": self.Output.dumps(),
            # Default to empty autoscaling
            "autoscaling": self.Trigger.AutoScaling.dumps(),
        }

        # @Deprecated dumping autoscaling and outputs at this level
        trigger = self.Trigger.get_trigger()
        try:
            if trigger.AutoScaling.autoscaling_config is not None:
                dump_config["autoscaling"] = trigger.AutoScaling.dumps()
        except AttributeError:
            pass

        try:
            if len(trigger.Output.resources) > 0:
                dump_config["outputs"] = trigger.Output.dumps()
        except AttributeError:
            pass

        return json.dumps(dump_config)



import pytest

from beam.utils.parse import compose_cpu, compose_memory


class TestParsers:
    def test_compose_cpu(self):
        assert compose_cpu(1) == "1000m"
        assert compose_cpu(1.999) == "1999m"
        assert compose_cpu("1000m") == "1000m"

    def test_compose_memory(self):
        assert compose_memory("10gi") == "10Gi"
        assert compose_memory("10 gi ") == "10Gi"
        assert compose_memory("10 Gi ") == "10Gi"

        # Raises if Gi is > 256
        with pytest.raises(ValueError):
            compose_memory("10000Gi")

        assert compose_memory("256mi") == "256Mi"
        assert compose_memory("256 mi ") == "256Mi"
        assert compose_memory("256 Mi ") == "256Mi"

        # Raises if Mi is < 128 or >= 1000
        with pytest.raises(ValueError):
            compose_memory("127Mi")

        with pytest.raises(ValueError):
            compose_memory("1000Mi")

        # Test invalid formats
        with pytest.raises(ValueError):
            compose_memory("1000ti")

        with pytest.raises(ValueError):
            compose_memory("1000.0")

        with pytest.raises(ValueError):
            compose_memory("1000")

        with pytest.raises(ValueError):
            compose_memory("2 gigabytes")



import beam
import json
import pytest

from beam.exceptions import BeamSerializationError
from beam.tests.utils import valid_reconstruction, generic_test_configs
from beam.types import Types


REQUIREMENTS_FILE_CONTENT = """

torch
numpy==1.19.5

scikit-learn

"""


class TestApp:
    def test_basic_app_reconstruction(self):
        for config in generic_test_configs:
            app = beam.App(**config)
            assert valid_reconstruction(app)

    def test_python_packages_from_requirements_file(self, tmp_path):
        for config in generic_test_configs:
            requirements_path = tmp_path / "requirements.txt"
            requirements_path.write_text(REQUIREMENTS_FILE_CONTENT)
            assert requirements_path.read_text() == REQUIREMENTS_FILE_CONTENT

            config["python_packages"] = str(requirements_path)

            app = beam.App(
                **config,
            )

            assert len(app.Spec.python_packages) == 3
            assert [
                "torch",
                "numpy==1.19.5",
                "scikit-learn",
            ] == app.Spec.python_packages

        assert valid_reconstruction(app)

    def test_valid_schedule_expressions(self):
        for config in generic_test_configs:
            app1 = beam.App(**config)

            app1.Trigger.Schedule(
                when="* * * * *",
                handler="method.py:run",
            )

            assert valid_reconstruction(app1)

            app1 = beam.App(**config)

            app1.Trigger.Schedule(
                when="every 5m",
                handler="method.py:run",
            )

            assert valid_reconstruction(app1)

            app2 = beam.App(**config)

            app2.Trigger.Schedule(
                when="every",
                handler="method.py:run",
            )

            with pytest.raises(BeamSerializationError):
                valid_reconstruction(app2)

    def test_multiple_triggers_should_fail_build(self):
        for config in generic_test_configs:
            app = beam.App(**config)

            app.Trigger.Schedule(
                when="* * * * *",
                handler="test.py:app",
            )

            with pytest.raises(ValueError):
                app.Trigger.TaskQueue(
                    inputs={
                        "some_input": beam.Types.String(),
                    },
                    handler="test.py:app",
                )

    def test_cpu_in_app_config(self):
        for config in generic_test_configs:
            config["cpu"] = 4
            app1 = beam.App(**config)

            config["cpu"] = "4000m"
            app2 = beam.App(**config)

            assert valid_reconstruction(app1)
            assert valid_reconstruction(app2)
            assert app1.dumps() == app2.dumps()

            config["cpu"] = 1
            app3 = beam.App(**config)

            config["cpu"] = "1000m"
            app4 = beam.App(**config)

            assert valid_reconstruction(app3)
            assert valid_reconstruction(app4)
            assert app3.dumps() == app4.dumps()

    def test_valid_rest_api(self):
        for config in generic_test_configs:
            app = beam.App(**config)

            schema = {
                "input1": beam.Types.String(),
                "input2": beam.Types.Float(),
                "input4": beam.Types.Boolean(),
                "input5": beam.Types.Json(),
            }

            app.Trigger.RestAPI(
                inputs=schema,
                outputs=schema,
                handler="method.py:run",
                loader="method.py:load",
                keep_warm_seconds=60,
                some_other_field="some_other_value",
            )

            assert valid_reconstruction(app)
            app_json = json.loads(app.dumps())
            assert app_json["triggers"][0]["inputs"] == Types.dump_schema(schema)
            assert app_json["triggers"][0]["outputs"] == Types.dump_schema(schema)
            assert app_json["triggers"][0]["handler"] == "method.py:run"
            assert app_json["triggers"][0]["loader"] == "method.py:load"
            assert app_json["triggers"][0]["keep_warm_seconds"] == 60
            assert "some_other_field" not in app_json["triggers"][0]

    def test_valid_webhook(self):
        for config in generic_test_configs:
            app_1 = beam.App(**config)
            app_2 = beam.App(**config)

            schema = {
                "input1": beam.Types.String(),
                "input2": beam.Types.Float(),
                "input4": beam.Types.Boolean(),
                "input5": beam.Types.Json(),
            }

            # Deprecate Webhook
            app_1.Trigger.Webhook(
                inputs=schema,
                outputs=schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )

            assert valid_reconstruction(app_1)
            app_json = json.loads(app_1.dumps())
            assert app_json["triggers"][0]["inputs"] == Types.dump_schema(schema)
            assert "outputs" not in app_json["triggers"][0]
            assert app_json["triggers"][0]["handler"] == "method.py:run"
            assert app_json["triggers"][0]["loader"] == "method.py:load"
            assert app_json["triggers"][0]["max_pending_tasks"] == 10
            assert "something_else" not in app_json["triggers"][0]

            # Use TaskQueue
            app_2.Trigger.TaskQueue(
                inputs=schema,
                outputs=schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )

            assert valid_reconstruction(app_2)
            app_json = json.loads(app_2.dumps())
            assert app_json["triggers"][0]["inputs"] == Types.dump_schema(schema)
            assert "outputs" not in app_json["triggers"][0]
            assert app_json["triggers"][0]["handler"] == "method.py:run"
            assert app_json["triggers"][0]["loader"] == "method.py:load"
            assert app_json["triggers"][0]["max_pending_tasks"] == 10
            assert "something_else" not in app_json["triggers"][0]

            assert app_1.dumps() == app_2.dumps()

    def test_valid_schedule(self):
        for config in generic_test_configs:
            app = beam.App(**config)

            app.Trigger.Schedule(
                when="* * * * *",
                handler="method.py:run",
                loader="method.py:load",
                something_else="some_other_value",
            )

            assert valid_reconstruction(app)
            app_json = json.loads(app.dumps())
            assert app_json["triggers"][0]["when"] == "* * * * *"
            assert app_json["triggers"][0]["handler"] == "method.py:run"
            assert "loader" not in app_json["triggers"][0]
            assert "something_else" not in app_json["triggers"][0]



from typing import Dict, Optional

from beam.base import Resource
from beam.serializer import RestAPITrigger, TaskQueueTrigger, CronJobTrigger
from beam.types import Types
from beam.configs.resource.autoscaling import AutoscalingResource
from beam.configs.resourceset.output import OutputResourceSet


class TriggerResourceMixin(Resource):
    @classmethod
    def from_config(cls, trigger: dict):
        inputs = {}
        outputs = {}

        if trigger.get("inputs") is not None:
            inputs = Types.load_schema(trigger.get("inputs"))

        if "inputs" in trigger:
            del trigger["inputs"]

        if trigger.get("outputs") is not None:
            outputs = Types.load_schema(trigger.get("outputs"))

        if "outputs" in trigger:
            del trigger["outputs"]

        return cls(**trigger, inputs=inputs, outputs=outputs)


class RestAPIResource(TriggerResourceMixin):
    def __init__(
        self,
        inputs: Dict[str, Types],
        handler: str,
        outputs: Optional[Dict[str, Types]] = None,
        loader: Optional[str] = None,
        keep_warm_seconds: Optional[int] = None,
        **_
    ):
        self.config = RestAPITrigger(
            inputs=inputs,
            outputs=outputs,
            handler=handler,
            loader=loader,
            keep_warm_seconds=keep_warm_seconds,
        )


class TaskQueueResource(TriggerResourceMixin):
    def __init__(
        self,
        inputs: Dict[str, Types],
        handler: str,
        loader: Optional[str] = None,
        max_pending_tasks: Optional[int] = 1000,
        callback_url: Optional[str] = None,
        **_
    ):
        self.config = TaskQueueTrigger(
            inputs=inputs,
            handler=handler,
            loader=loader,
            max_pending_tasks=max_pending_tasks,
            callback_url=callback_url,
        )

        self.AutoScaling = AutoscalingResource()
        self.Output = OutputResourceSet()


class ScheduledJobResource(TriggerResourceMixin):
    def __init__(self, when: str, handler: str, **_):
        self.config = CronJobTrigger(when=when, handler=handler)
        self.Output = OutputResourceSet()



import dataclasses

from typing import List, Optional, Union
from marshmallow import validate
from marshmallow_dataclass import dataclass

from beam.types import PythonVersion, OutputType, MountType, AutoscalingType, GpuType
from beam.base import BaseSerializer, BaseTriggerSerializer
from beam.utils.parse import compose_cpu, compose_memory
from beam import validators


@dataclass
class AppSpecV1Configuration(BaseSerializer):
    name: str = validators.field(validate.Length(max=128))
    cpu: str = validators.field(validators.SerializerMethod(compose_cpu))
    gpu: str = validators.field(
        validate.OneOf(choices=[gpu[1] for gpu in GpuType.Types])
    )
    memory: str = validators.field(validators.SerializerMethod(compose_memory))
    python_version: str = validators.field(
        validate.OneOf(choices=[version[1] for version in PythonVersion.Types])
    )
    python_packages: Union[str, List[str]]
    workspace: str


@dataclass
class AppSpecV2Configuration(BaseSerializer):
    name: str = validators.field(validate.Length(max=128))
    cpu: str = validators.field(validators.SerializerMethod(compose_cpu))
    gpu: str = validators.field(
        validate.OneOf(choices=[gpu[1] for gpu in GpuType.Types])
    )
    memory: str = validators.field(validators.SerializerMethod(compose_memory))
    python_version: str = validators.field(
        validate.OneOf(choices=[version[1] for version in PythonVersion.Types])
    )
    python_packages: Union[str, List[str]]
    commands: Union[str, List[str]]
    workspace: str


@dataclass
class RestAPITrigger(BaseTriggerSerializer):
    inputs: validators.TypeSerializerDict(required=True)
    outputs: validators.TypeSerializerDict()
    handler: str = validators.field(validators.IsFileMethod())
    loader: Optional[str] = validators.field(validators.IsFileMethod())
    keep_warm_seconds: Optional[int]
    trigger_type: str = "rest_api"


@dataclass
class CronJobTrigger(BaseTriggerSerializer):
    when: str = validators.field(validators.IsValidCronOrEvery())
    handler: str = validators.field(validators.IsFileMethod())
    trigger_type: str = "cron_job"


@dataclass
class TaskQueueTrigger(BaseTriggerSerializer):
    inputs: validators.TypeSerializerDict(required=True)
    handler: str = validators.field(validators.IsFileMethod())
    loader: Optional[str] = validators.field(validators.IsFileMethod())
    callback_url: Optional[str] = validators.field(validators.IsValidURL())
    max_pending_tasks: Optional[int] = 1000
    trigger_type: str = "webhook"


@dataclass
class FileConfiguration(BaseSerializer):
    path: str
    name: str
    output_type: str = validators.field(
        validate.OneOf(choices=[version[1] for version in OutputType.Types])
    )


@dataclass
class VolumeConfiguration(BaseSerializer):
    name: str
    app_path: str
    mount_type: str = validators.field(
        validate.OneOf(choices=[t[1] for t in MountType.Types])
    )
    checksum: str = None
    local_path: str = None


@dataclass
class AutoscalingMaxRequestLatencyConfiguration(BaseSerializer):
    desired_latency: float
    max_replicas: int
    autoscaling_type: str = dataclasses.field(
        metadata={
            "validate": validate.OneOf(choices=[t[1] for t in AutoscalingType.Types])
        }
    )



import json
import beam

from beam.tests.utils import valid_reconstruction, generic_test_configs


class TestAutoScaling:
    # Should be removed once we fully remove autoscaling at root level of config json
    def test_old_autoscaling(self):
        for config in generic_test_configs:
            app = beam.App(
                **config,
            )

            app.Trigger.AutoScaling.MaxRequestLatency(
                desired_latency=10, max_replicas=100
            )

            assert valid_reconstruction(app)
            app_json = json.loads(app.dumps())
            assert app_json["autoscaling"]["max_replicas"] == 100
            assert (
                app_json["autoscaling"]["autoscaling_type"]
                == beam.types.AutoscalingType.MaxRequestLatency
            )
            assert app_json["autoscaling"]["desired_latency"] == 10

    # Should be removed once we fully remove autoscaling at root level of config json
    def test_old_autoscaling_matches_current_autoscaling(self):
        for config in generic_test_configs:
            app1 = beam.App(
                **config,
            )

            schema = {
                "input1": beam.Types.String(),
                "input2": beam.Types.Float(),
                "input4": beam.Types.Boolean(),
                "input5": beam.Types.Json(),
            }

            # Deprecate Webhook
            app1.Trigger.TaskQueue(
                inputs=schema,
                outputs=schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )
            app1.Trigger.AutoScaling.MaxRequestLatency(
                desired_latency=10, max_replicas=100
            )

            assert valid_reconstruction(app1)
            app_json_1 = json.loads(app1.dumps())
            assert app_json_1["autoscaling"]["max_replicas"] == 100
            assert (
                app_json_1["autoscaling"]["autoscaling_type"]
                == beam.types.AutoscalingType.MaxRequestLatency
            )
            assert app_json_1["autoscaling"]["desired_latency"] == 10

            app2 = beam.App(
                **config,
            )

            tq = app2.Trigger.TaskQueue(
                inputs=schema,
                outputs=schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )

            tq.AutoScaling.MaxRequestLatency(desired_latency=10, max_replicas=100)

            assert valid_reconstruction(app2)
            app_json_2 = json.loads(app2.dumps())
            assert app_json_2 == app_json_1



from typing import Dict, Optional, Union
from warnings import warn

from beam.base import ResourceSet
from beam.configs.resource import (
    TaskQueueResource,
    RestAPIResource,
    ScheduledJobResource,
    AutoscalingResource,
)
from beam.types import Types


class TriggerType:
    Webhook = "webhook"
    RestAPI = "rest_api"
    CronJob = "cron_job"


class TriggerResourceSet(ResourceSet):
    def __init__(self) -> None:
        super().__init__()

        # @Deprecate AutoScaling in TriggerResourceSet
        # Opt into using Autoscaling through TaskQueueResource instead
        self.AutoScaling = AutoscalingResource()

    def _validate_trigger_groupings(self):
        """
        NOTE: For the time being, the Beam APP can only accept one trigger during the alpha
        stages. Later we will allow multiple trigger types for webhooks (Slack, Twitter, etc)
        """
        if len(self.resources) > 1:
            raise ValueError("App can only have 1 trigger at a time")

    def Webhook(
        self,
        inputs: Dict[str, Types],
        handler: str,
        loader: Optional[str] = None,
        max_pending_tasks: Optional[int] = 1000,
        callback_url: Optional[str] = None,
        **_,
    ):
        """
        Arguments:
            handler: specify what method in which file to use as the entry point
            loader: specify the method that will load the data into the handler (optional)
            inputs: dictionary specifying how to serialize/deserialize input arguments
            max_pending_tasks (optional): maximum number of pending tasks in the queue, if exceeded, the webhook will prevent further items from being enqueued

        """
        warn(
            "The 'Webhook' trigger is deprecated and is replaced by 'TaskQueue'. Please use 'TaskQueue' instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self.TaskQueue(
            inputs=inputs,
            handler=handler,
            loader=loader,
            max_pending_tasks=max_pending_tasks,
            callback_url=callback_url,
        )

    def TaskQueue(
        self,
        inputs: Dict[str, Types],
        handler: str,
        loader: Optional[str] = None,
        max_pending_tasks: Optional[int] = 1000,
        callback_url: Optional[str] = None,
        **_,
    ):
        """
        Arguments:
            handler: specify what method in which file to use as the entry point
            inputs: dictionary specifying how to serialize/deserialize input arguments
            max_pending_tasks (optional): maximum number of pending tasks in the queue, if exceeded, the webhook will prevent further items from being enqueued

        """
        resource = TaskQueueResource(
            inputs=inputs,
            handler=handler,
            loader=loader,
            max_pending_tasks=max_pending_tasks,
            callback_url=callback_url,
        )

        self.resources.append(resource)
        self._validate_trigger_groupings()

        return resource

    def Schedule(self, when: str, handler: str, **_):
        """
        Arguments:
            handler: specify what method in which file to use as the entry point
            when: CRON string to indicate the schedule in which the job is to run
                - https://en.wikipedia.org/wiki/Cron \
        """
        resource = ScheduledJobResource(when=when, handler=handler)
        self.resources.append(resource)
        self._validate_trigger_groupings()

        return resource

    def RestAPI(
        self,
        inputs: Dict[str, Types],
        handler: str,
        outputs: Optional[Dict[str, Types]] = None,
        loader: Optional[str] = None,
        keep_warm_seconds: Optional[int] = None,
        **_,
    ):
        """
        Arguments:
            handler: specify what method in which file to use as the entry point
            loader: specify the method that will load the data into the handler (optional)
            inputs: dictionary specifying how to serialize/deserialize input arguments
            outputs: dictionary specifying how to serialize/deserialize return values
            keep_warm_seconds: how long the app should stay warm after the last request
        """
        resource = RestAPIResource(
            inputs=inputs,
            handler=handler,
            outputs=outputs,
            loader=loader,
            keep_warm_seconds=keep_warm_seconds,
        )

        self.resources.append(resource)
        self._validate_trigger_groupings()

        return resource

    def get_trigger(
        self,
    ) -> Union[ScheduledJobResource, TaskQueueResource, RestAPIResource]:
        self._validate_trigger_groupings()

        if len(self.resources) == 0:
            return None

        return self.resources[0]

    def dumps(self):
        # To make this backwards compatible in the future after switching back to
        # multiple triggers, we will make this a list that currently will only have 1 trigger
        self._validate_trigger_groupings()
        triggers = []

        for r in self.resources:
            triggers.append(r.dumps())

        return triggers

    def from_config(self, triggers):
        if triggers is None:
            return

        for t in triggers:
            trigger_type = t.get("trigger_type")

            if trigger_type == TriggerType.Webhook:
                self.resources.append(TaskQueueResource.from_config(t))
            elif trigger_type == TriggerType.CronJob:
                self.resources.append(ScheduledJobResource.from_config(t))
            elif trigger_type == TriggerType.RestAPI:
                self.resources.append(RestAPIResource.from_config(t))
            else:
                raise ValueError(
                    f"Found an unknown trigger type in config: {trigger_type}"
                )



import json
import beam

from beam.types import PythonVersion

generic_test_configs = [
    {
        "name": "some_app",
        "cpu": "4000m",
        "memory": "128mi",
        "gpu": "A10G",
        "python_packages": ["pytorch"],
        "python_version": PythonVersion.Python37,
        "workspace": "./some_folder",
    },
    {
        "name": "some_app",
        "cpu": 1,
        "memory": "128mi",
        "gpu": "A10G",
    },
    {
        "name": "some_app",
        "cpu": 1,
        "memory": "1Gi",
    },
]


def valid_reconstruction(app: beam.App):
    dumped_app = app.dumps()
    dumped_json = json.dumps(dumped_app)
    app_from_dumped_config = beam.app.App.from_config(json.loads(dumped_json))
    return app_from_dumped_config.dumps() == dumped_app



def hello_world(**inputs):
    print("After deploying, you'll see these logs in the web dashboard 👀")
    return {"response": inputs["text"]}


if __name__ == "__main__":
    text = "Testing 123"
    hello_world(text=text)



import os
import importlib
import traceback
import importlib
import json

from http import HTTPStatus
from beam.app import App
from beam.types import Types, InvalidPayloadException, InvalidTypeException


class MockAPI:
    """
    Arguments:
        handler: specify what method in which file to use as the entry point
        loader:
    """

    def __init__(self, app: App) -> None:
        self.app = app
        self._load_trigger()
        self._load_handler()

        if self.loader is not None and self.loader != "":
            self.loader_output = self._run_loader()

    def _load_trigger(self) -> None:
        self.trigger = self.app.Trigger.get_trigger().config.validate_and_dump()
        self.handler = self.trigger.get("handler", None)
        self.loader = self.trigger.get("loader")
        self.loader_output = None

    # Run loader
    def _run_loader(self):
        module_file, func = self.loader.split(":")
        module = os.path.splitext(module_file)[0]

        try:
            loader_module = importlib.import_module(module)
            self.loader_method = getattr(loader_module, func)
            return self.loader_method()
        except BaseException:
            raise Exception(f"Unable to run loader: {traceback.format_exc()}")

    # Load handler method
    def _load_handler(self) -> None:
        inputs = self.trigger.get("inputs", None)

        if inputs is not None:
            self.inputs = Types.load_schema(inputs)
        else:
            self.inputs = {}

        outputs = self.trigger.get("outputs", None)
        if outputs is not None:
            self.outputs = Types.load_schema(outputs)

        module_file, func = self.handler.split(":")
        module = os.path.splitext(module_file)[0]

        try:
            # instantiate handler and inputs / outputs
            handler_module = importlib.import_module(module)
            self.handler_method = getattr(handler_module, func)
        except BaseException:
            raise Exception("Unable to load handler")

    def call(self, **_input):
        deserialization_errors = []
        deserialized_payload = {}
        serialized_output = {}

        # serialize and validate inputs
        try:
            deserialized_payload = Types.deserialize(_input, self.inputs)
        except (InvalidPayloadException, InvalidTypeException) as exc:
            deserialization_errors = exc.errors

        if deserialization_errors:
            return {"errors": deserialization_errors}, HTTPStatus.BAD_REQUEST

        # If the user ran a loader prior to starting the api, pass in the object they returned
        if self.loader_output is not None:
            deserialized_payload["loader_output"] = self.loader_output

        try:
            output = self.handler_method(**deserialized_payload)
        except BaseException:
            return {
                "errors": [traceback.format_exc()]
            }, HTTPStatus.INTERNAL_SERVER_ERROR

        # serialize and validate outputs
        try:
            serialized_output = Types.serialize(output, self.outputs)
        except (
            InvalidPayloadException,
            InvalidTypeException,
            BaseException,
        ) as exc:
            serialization_errors = getattr(exc, "errors", [])
            return {"errors": serialization_errors}, HTTPStatus.INTERNAL_SERVER_ERROR

        return serialized_output, HTTPStatus.OK



import json
import beam
import pytest

from beam.tests.utils import valid_reconstruction, generic_test_configs
from beam.exceptions import BeamSerializationError


class TestMounts:
    def test_persistent_volume(self):
        for config in generic_test_configs:
            app = beam.App(
                **config,
            )

            app.Mount.PersistentVolume(
                path="/some/path",
                name="some_name",
            )

            assert valid_reconstruction(app)

            app_json = json.loads(app.dumps())
            assert app_json.get("mounts") is not None
            assert app_json.get("mounts")[0].get("app_path") == "/some/path"
            assert app_json.get("mounts")[0].get("name") == "some_name"
            assert (
                app_json.get("mounts")[0].get("mount_type")
                == beam.types.MountType.Persistent
            )

    def test_persistent_volume_missing_values(self):
        for config in generic_test_configs:
            app = beam.App(
                **config,
            )

            app.Mount.PersistentVolume(
                name="some_name",
            )

            with pytest.raises(BeamSerializationError):
                app.dumps()

            app = beam.App(
                **config,
            )

            with pytest.raises(TypeError):
                app.Mount.PersistentVolume(
                    path="/some/path",
                )

    def test_shared_volume_missing_values(self):
        for config in generic_test_configs:
            app = beam.App(
                **config,
            )

            app.Mount.SharedVolume(
                name="some_name",
            )

            with pytest.raises(BeamSerializationError):
                app.dumps()

            app = beam.App(
                **config,
            )

            with pytest.raises(TypeError):
                app.Mount.SharedVolume(
                    path="/some/path",
                )

    def test_shared_volume(self):
        for config in generic_test_configs:
            app = beam.App(
                **config,
            )

            app.Mount.SharedVolume(
                path="/some/path",
                name="some_name",
            )

            assert valid_reconstruction(app)

            app_json = json.loads(app.dumps())
            assert app_json.get("mounts") is not None
            assert app_json.get("mounts")[0].get("app_path") == "/some/path"
            assert app_json.get("mounts")[0].get("name") == "some_name"
            assert (
                app_json.get("mounts")[0].get("mount_type")
                == beam.types.MountType.Shared
            )



import json
import beam

from beam.tests.utils import valid_reconstruction, generic_test_configs


test_schema = {
    "input1": beam.Types.String(),
    "input2": beam.Types.Float(),
    "input4": beam.Types.Boolean(),
    "input5": beam.Types.Json(),
}


class TestOutputs:
    def _verify_output(self, app, name, path, output_type):
        assert valid_reconstruction(app)

        app_json = json.loads(app.dumps())
        assert app_json.get("outputs") is not None
        assert app_json.get("outputs")[0].get("path") == path
        assert app_json.get("outputs")[0].get("name") == name
        assert app_json.get("outputs")[0].get("output_type") == output_type

        return app_json

    def test_add_files_webhook(self):
        for config in generic_test_configs:
            # Old way of adding outputs
            app1 = beam.App(
                **config,
            )

            app1.Trigger.TaskQueue(
                inputs=test_schema,
                outputs=test_schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )

            app1.Output.File(
                path="/some/path",
                name="some_name",
            )

            assert valid_reconstruction(app1)

            app_json_1 = self._verify_output(
                app1, "some_name", "/some/path", beam.types.OutputType.File
            )

            # New way of adding outputs
            app2 = beam.App(
                **config,
            )

            trigger = app2.Trigger.TaskQueue(
                inputs=test_schema,
                outputs=test_schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )

            trigger.Output.File(
                path="/some/path",
                name="some_name",
            )

            app_json_2 = self._verify_output(
                app2, "some_name", "/some/path", beam.types.OutputType.File
            )

            assert app_json_1 == app_json_2

    def test_add_directory_webhook(self):
        for config in generic_test_configs:
            app1 = beam.App(
                **config,
            )

            app1.Trigger.TaskQueue(
                inputs=test_schema,
                outputs=test_schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )

            app1.Output.Dir(
                path="/some/path",
                name="some_name",
            )

            app_json_1 = self._verify_output(
                app1, "some_name", "/some/path", beam.types.OutputType.Directory
            )

            # New way of adding outputs
            app2 = beam.App(
                **config,
            )

            trigger = app2.Trigger.TaskQueue(
                inputs=test_schema,
                outputs=test_schema,
                handler="method.py:run",
                loader="method.py:load",
                max_pending_tasks=10,
                something_else="some_other_value",
            )

            trigger.Output.Dir(
                path="/some/path",
                name="some_name",
            )

            app_json_2 = self._verify_output(
                app2, "some_name", "/some/path", beam.types.OutputType.Directory
            )

            assert app_json_1 == app_json_2

    def test_add_file_scheduled_job(self):
        for config in generic_test_configs:
            app1 = beam.App(
                **config,
            )

            app1.Trigger.Schedule(
                when="every 5m",
                handler="method.py:run",
            )

            app1.Output.File(
                path="/some/path",
                name="some_name",
            )

            app_json_1 = self._verify_output(
                app1, "some_name", "/some/path", beam.types.OutputType.File
            )

            # New way of adding outputs
            app2 = beam.App(
                **config,
            )

            trigger = app2.Trigger.Schedule(
                when="every 5m",
                handler="method.py:run",
            )

            trigger.Output.File(
                path="/some/path",
                name="some_name",
            )

            app_json_2 = self._verify_output(
                app2, "some_name", "/some/path", beam.types.OutputType.File
            )

            assert app_json_1 == app_json_2

    def test_add_directory_scheduled_job(self):
        for config in generic_test_configs:
            app1 = beam.App(
                **config,
            )

            app1.Trigger.Schedule(
                when="every 5m",
                handler="method.py:run",
            )

            app1.Output.Dir(
                path="/some/path",
                name="some_name",
            )

            app_json_1 = self._verify_output(
                app1, "some_name", "/some/path", beam.types.OutputType.Directory
            )

            # New way of adding outputs
            app2 = beam.App(
                **config,
            )

            trigger = app2.Trigger.Schedule(
                when="every 5m",
                handler="method.py:run",
            )

            trigger.Output.Dir(
                path="/some/path",
                name="some_name",
            )

            app_json_2 = self._verify_output(
                app2, "some_name", "/some/path", beam.types.OutputType.Directory
            )

            assert app_json_1 == app_json_2



import os
from pathlib import Path

DEPLOYMENT_READY_PATH = "/workspace/READY"


def is_local():
    app_id = os.environ.get("APP_ID")
    app_task_name = os.environ.get("APP_TASK_NAME")

    if app_id is None or app_task_name is None:
        return True


def checkpoint():
    if is_local():
        return

    Path(DEPLOYMENT_READY_PATH).touch()
